package swingworkerdemo.fp;

import java.util.EventListener;

/**
 * The Functional Part of the demo:
 * a slow implementation of the Fibonacci function.
 * This calculation is observable via the ProgressListener interface,
 * on which it reports progress on a scale from 0 to 100.
 *
 * @author Tom Verhoeff (TU/e)
 */
public class Calculation {

    /** Lower bound on arguments for method {@code fib}. */
    public static final int LOWER_BOUND = 0;

    /**
     * Upper bound on arguments for method {@code fib}.
     * fib(47) does not fit in a long;
     * this also provides some protection in case the calculation
     * is not run in a separate thread and cannot be aborted.
     */
    public static final int UPPER_BOUND = 45;

    /** Progress listener interface. */
    public static interface ProgressListener extends EventListener {

        /**
         * Reports progress.
         *
         * @param progress  the progress indicator
         * @pre {@code 0 <= progress <= 100}
         */
        void reportProgress(int progress);

    }

    /** Flag to abort the calculation. */
    private boolean abort;

    /** Number of calls of the recursive method. */
    private long count;

    /**
     * Expected number of calls of the recursive method:
     * 2 * fib(n + 1) - 1.
     */
    private long expectedCount;

    /** Step size for progress reporting. */
    private long step;

    /** Progress listener. */
    private ProgressListener listener;

    /**
     * Determines whether a given number is a valid argument for
     * method {@code fib}.
     *
     * @param n  the argument to validate
     * @return  {@code LOWER_BOUND <= n && n <= UPPER_BOUND}
     */
    public static boolean validArgument(final int n) {
        return LOWER_BOUND <= n && n <= UPPER_BOUND;
    }

    /**
     * Returns the call count.
     *
     * @return number of times {@code fib} was called since {@code resetCount}.
     */
    public long getCount() {
        return this.count;
    }

    /**
     * Resets the call count.
     *
     * @modifies {@code this}
     */
    public void resetCount() {
        this.count = 0;
    }

    /**
     * Requests the calculation to be aborted (at a convenient moment).
     */
    public void abort() {
        this.abort = true;
    }

    /**
     * Sets the progress listener.
     *
     * @param listener  the progress listener
     */
    public void setListener(final ProgressListener listener) {
        this.listener = listener;
    }

    /**
     * Returns the {@code n}-th Fibonacci number.
     *
     * @param n  the index, starting at 0
     * @return the {@code n}-th Fibonacci number
     * @pre   {@code validArgument(n)}
     * @modifies {@code this}
     */
    public long fib(final int n) {
        if (! validArgument(n)) {
            throw new IllegalArgumentException(
                    "fib(" + n + ") violates @pre: argument out of range");
        }

        this.count = 0;
        this.expectedCount = 2 * fibLinear(n + 1) - 1;
        this.step = expectedCount / 100;
        this.abort = false;
        long result = -1;
        try {
            result = fibRecursive(n);
        } catch (InterruptedException e) {
            // aborted, ignore exception
        }
        return result;
    }

    /**
     * Returns the {@code n}-th Fibonacci number.
     * Inefficient implementation (exponential in {@code n}).
     * The calculation can be interrupted by setting {@code abort}.
     *
     * @param n  the index, starting at 0
     * @return the {@code}-th Fibonacci number
     * @pre   {@code validArgument(n)}
     * @modifies {@code this}
     */
    private long fibRecursive(final int n) throws InterruptedException {
        if (abort) {
            throw new InterruptedException("Aborted by user");
        }

        ++ count;
        if (listener != null && count % step == 0) {
            listener.reportProgress(
                    (int)Math.round(count * 100.0 / expectedCount));
        }

        if (n <= 1) {
            return n;
        } else {
            return fibRecursive(n - 1) + fibRecursive(n - 2);
        }
    }

    /**
     * Returns the {@code n}-th Fibonacci number.
     * Linear implementation.
     *
     * @param n  the index, starting at 0
     * @return the {@code}-th Fibonacci number
     * @pre   {@code 0 <= n}
     */
    private long fibLinear(final int n) {
        long a = 0;
        long b = 1;
        int k = 0;
        // invariants:
        //   0 <= k <= n
        //   a == fib(k)
        //   b == fib(k + 1)

        while (k != n) {
            final long c = a + b; // c == fib(k + 2)
            a = b;
            b = c;
            ++ k;
        }

        return a;
    }

}
